diff -uNr trunk/encoder/encoder.c trunk-new/encoder/encoder.c
--- trunk/encoder/encoder.c	Tue Nov 20 02:57:28 2007
+++ trunk-new/encoder/encoder.c	Wed Jan  2 22:27:03 2008
@@ -615,7 +615,7 @@
     x264_reduce_fraction( &h->param.i_fps_num, &h->param.i_fps_den );
 
     /* Init x264_t */
-    h->i_frame = 0;
+    h->i_frame = -1;
     h->i_frame_num = 0;
     h->i_idr_pic_id = 0;
 
@@ -942,9 +942,6 @@
 {
     int i;
 
-    if( h->fdec->i_frame >= 0 )
-        h->i_frame++;
-
     if( !h->fdec->b_kept_as_ref )
     {
         if( h->param.i_threads > 1 )
@@ -1314,6 +1311,7 @@
         }
     }
 
+    h->i_frame++;
     if( h->frames.current[0] == NULL )
     {
         int bframes = 0;
@@ -1775,6 +1773,17 @@
         if( h->thread[i]->b_thread_active )
             x264_pthread_join( h->thread[i]->thread_handle, NULL );
     }
+
+    if( h->param.i_threads > 1)
+    {
+        x264_t *thread_prev;
+
+        thread_prev = h->thread[ h->i_thread_phase % h->param.i_threads ];
+        x264_thread_sync_ratecontrol( h, thread_prev, h );
+        x264_thread_sync_ratecontrol( thread_prev, thread_prev, h );
+        h->i_frame = thread_prev->i_frame + 1 - h->param.i_threads;
+    }
+    h->i_frame++;
 
 #ifdef DEBUG_BENCHMARK
     x264_log( h, X264_LOG_INFO,
diff -uNr trunk/encoder/ratecontrol.c trunk-new/encoder/ratecontrol.c
--- trunk/encoder/ratecontrol.c	Wed Sep 12 02:42:22 2007
+++ trunk-new/encoder/ratecontrol.c	Wed Jan  2 22:25:50 2008
@@ -358,21 +358,19 @@
             x264_log( h, X264_LOG_WARNING, "2nd pass has fewer frames than 1st pass (%d vs %d)\n",
                       h->param.i_frame_total, rc->num_entries );
         }
-        if( h->param.i_frame_total > rc->num_entries + h->param.i_bframe )
+        if( h->param.i_frame_total > rc->num_entries + h->param.i_bframe + X264_THREAD_MAX )
         {
             x264_log( h, X264_LOG_ERROR, "2nd pass has more frames than 1st pass (%d vs %d)\n",
                       h->param.i_frame_total, rc->num_entries );
             return -1;
         }
 
-        /* FIXME: ugly padding because VfW drops delayed B-frames */
-        rc->num_entries += h->param.i_bframe;
-
-        rc->entry = (ratecontrol_entry_t*) x264_malloc(rc->num_entries * sizeof(ratecontrol_entry_t));
-        memset(rc->entry, 0, rc->num_entries * sizeof(ratecontrol_entry_t));
+        /* FIXME: ugly padding because VfW drops delayed B-frames and frames delayed by multithreading */
+        rc->entry = (ratecontrol_entry_t*) x264_malloc((rc->num_entries + h->param.i_bframe + X264_THREAD_MAX) * sizeof(ratecontrol_entry_t));
+        memset(rc->entry, 0, (rc->num_entries + h->param.i_bframe + X264_THREAD_MAX) * sizeof(ratecontrol_entry_t));
 
         /* init all to skipped p frames */
-        for(i=0; i<rc->num_entries; i++){
+        for(i=0; i < rc->num_entries + h->param.i_bframe + X264_THREAD_MAX; i++){
             ratecontrol_entry_t *rce = &rc->entry[i];
             rce->pict_type = SLICE_TYPE_P;
             rce->qscale = rce->new_qscale = qp2qscale(20);
@@ -382,7 +380,7 @@
 
         /* read stats */
         p = stats_in;
-        for(i=0; i < rc->num_entries - h->param.i_bframe; i++){
+        for(i=0; i < rc->num_entries; i++){
             ratecontrol_entry_t *rce;
             int frame_number;
             char pict_type;
@@ -397,7 +395,7 @@
             }
             e = sscanf(p, " in:%d ", &frame_number);
 
-            if(frame_number < 0 || frame_number >= rc->num_entries)
+            if(frame_number < 0 || frame_number >= rc->num_entries + h->param.i_bframe + X264_THREAD_MAX)
             {
                 x264_log(h, X264_LOG_ERROR, "bad frame number (%d) at stats line %d\n", frame_number, i);
                 return -1;
@@ -602,7 +600,7 @@
     if( rc->p_stat_file_out )
     {
         fclose( rc->p_stat_file_out );
-        if( h->i_frame >= rc->num_entries - h->param.i_bframe )
+        if( h->i_frame >= rc->num_entries )
             if( rename( rc->psz_stat_file_tmpname, h->param.rc.psz_stat_out ) != 0 )
             {
                 x264_log( h, X264_LOG_ERROR, "failed to rename \"%s\" to \"%s\"\n",
@@ -656,7 +654,7 @@
     if( h->param.rc.b_stat_read )
     {
         int frame = h->fenc->i_frame;
-        assert( frame >= 0 && frame < rc->num_entries );
+        assert( frame >= 0 && frame < rc->num_entries + h->param.i_bframe + X264_THREAD_MAX );
         rce = h->rc->rce = &h->rc->entry[frame];
 
         if( h->sh.i_type == SLICE_TYPE_B
@@ -833,7 +831,7 @@
     x264_ratecontrol_t *rc = h->rc;
     if( h->param.rc.b_stat_read )
     {
-        if( frame_num >= rc->num_entries )
+        if( frame_num >= rc->num_entries + h->param.i_bframe + X264_THREAD_MAX )
         {
             /* We could try to initialize everything required for ABR and
              * adaptive B-frames, but that would be complicated.